/*
	EmacsEdit.cpp  --- Jonathan Arnold   December 10 1997

	Main module for interfacing Borland's IDE with GNU Emacs on Win9x / WinNT

	This file is not part of GNU Emacs.

	Copyright(C) 1999 Jonathan Arnold (jdarnold@buddydog.org)

	This module implements TEmacsEdit and TEmacsNotifier.  TEmacsEdit is a
	class that interfaces to Borland IDE's ToolServices, as shown in the sample
	source file ToolIntf.pas.  It also creates a TEmacsClientRead object,
	which does the talking to GNUSERV via mailslots.  TEmacsNotifier is the
	callback used by the IDE to tell EmacsEdit when things are happening.

	Thursday, September 24 1998 : We don't do notifiers any more
    And I did lots of cleanup, including adding in Release calls for
		the Interfaces I got
*/

#include <Windows.hpp>
#include <ExptIntf.hpp>
#include <EditIntf.hpp>

#include "EmacsEdit.h"
 
#include "gnuserv.h"
#include "EmacsClientRead.h"

USERES("EmacsEdit.res");
USEUNIT("gnuslib.cpp");
USEUNIT("EmacsClientRead.cpp");
//---------------------------------------------------------------------------
HINSTANCE instance;
TEmacsEdit* ee;
TIToolServices* toolServices;

#define OutputDebugString //

// If you want to add the notifier back in:
//#define DO_NOTIFIER 1

// If you want to use GNUCLIENT to open files from the IDE in the editor
//#define USE_GNUCLIENT 1

//---------------------------------------------------------------------------
TEmacsEdit::TEmacsEdit( void ) : TIExpert()
{
	OutputDebugString( "TEmacsEdit constructor 1\n" );
	
	clientID = GetCurrentProcessId();

	emacsReader = 0;
	emacsNotifier = 0;
	
	// Double check that emacs is around
	serverSlot = ConnectToMailslot(GNUSERV_DEFAULT_HOST, FALSE, 0);
	if (INVALID_HANDLE_VALUE == serverSlot)
		return;
	DisconnectFromMailslot( serverSlot );

	clientMailSlot = CreateGnuServSlot( true, clientID );

	// create mailslot to receive results (EmacsClientReader uses it)
	emacsReader = new EmacsClientRead(this,true);

	emacsReader->SetMailslot( clientMailSlot );

	emacsReader->Resume();

#if DO_NOTIFIER
	OutputDebugString( "TEmacsEdit constructor 5\n" );
	emacsNotifier = new TEmacsNotifier();
	OutputDebugString( "TEmacsEdit constructor 6\n" );
	if ( emacsNotifier )
	{ // okay
		emacsNotifier->SetEditor( this );
		toolServices->AddNotifier( emacsNotifier );
	}
#endif
	
	AnsiString ci("(borlandide-init ");
	ci += (int)clientID;
	ci += ")";

	EmacsSend( ci );
}

bool TEmacsEdit::IsValid()
{
	bool valid = emacsReader && emacsReader->IsValid();
	
#if DO_NOTIFIER
	valid = valid && emacsNotifier;
#endif
	
	OutputDebugString( "IsValid :" );
	if (valid)
		OutputDebugString( "True\n" );
	else
		OutputDebugString( "False\n" );

 	return valid;
}

__fastcall TEmacsEdit::~TEmacsEdit()
{
	OutputDebugString( "Exiting TEmacsEdit\n" );
	
	if ( emacsReader )
	{
		// Tell emacs we're done
		EmacsSend("(borlandide-exit)");
		Sleep(2000);
	
		emacsReader->Terminate();
		DisconnectFromMailslot(clientMailSlot);
	}

	if ( emacsNotifier )
	{
		toolServices->RemoveNotifier( emacsNotifier );
		delete emacsNotifier;
	}
}

void TEmacsEdit::OpenFile( AnsiString fname )
{
#if defined(USE_GNUCLIENT)
	AnsiString cmdline("gnuclientw -F ");

	cmdline += fname;
	
	MessageBox( NULL, cmdline.c_str(), "OpenFile", MB_OK );
	
	WinExec( cmdline.c_str(), SW_HIDE );
#elif 0
	// Then send a lisp command
	AnsiString basecommand("(server-edit-files-quickly '((1 . \"");

	char* path = fname.c_str();

	// Make the drive letter be lower-case
	*path = (char)tolower(*path);
	do
	{ // make \ be / for Emacs
		if ( *path == '\\' )
			*path = '/';
	} while ( *++path );
	
	basecommand += fname;
	basecommand += "\")))";
	
	EmacsSend( basecommand );
#endif
}

bool TEmacsEdit::SendString( AnsiString msg )
{
	OutputDebugString( "Sendstring: \'" );
	OutputDebugString( msg.c_str() );
	OutputDebugString( "\'\n" );
	
	DWORD byteswritten;
	bool rval = false;

	if (!WriteFile(serverSlot, msg.c_str(), (DWORD)msg.Length() + 1,
			&byteswritten, (LPOVERLAPPED)NULL)) {
		MessageBox(NULL, "Unable to write to Emacs", "Problem", MB_OK);
	}
	else
		rval = true;

	return rval;
}

void TEmacsEdit::EmacsSend( AnsiString command )
{
	OutputDebugString( command.c_str() );
	OutputDebugString( "\n" );
	
	// Try to connect to the send mailslot
	serverSlot = ConnectToMailslot(GNUSERV_DEFAULT_HOST, FALSE, 0);
	if (INVALID_HANDLE_VALUE == serverSlot)
		return;

	AnsiString clientIDstr("C:");
	clientIDstr += AnsiString((int)clientID) + " ";
	if (!SendString(clientIDstr.c_str() ))
		return;

	SendString(command.c_str() );
	SendString(EOT_STR);

	DisconnectFromMailslot( serverSlot );
}

void TEmacsEdit::MsgFromEmacs( char* msg )
{
	OutputDebugString( "MessageFromEmacs: " );
	OutputDebugString( msg );
	OutputDebugString( "\n" );

	if ( strncmp(msg,"saved ",6) == 0 )
	{ // saved a file, so keep it from being complained about in the editor
		char* filename = strchr( msg, ' ' ) + 1;
		char *ch;
		
		for ( ch = filename; *ch; ch++ )
		{ // Fix up path separators
			if ( *ch == '/' )
				*ch = '\\';
		}

		// And uppercase the drive letter 
		*filename = (char)toupper((*filename));

		AnsiString fname(filename);

		if ( !toolServices->IsFileOpen( fname ) )
		{ // let's see if the short name works
			char short_fname[MAX_PATH];
			
			if ( ::GetShortPathName( fname.c_str(), short_fname, sizeof(short_fname)) != 0 )
			{ // worked, so try that as the fname
				OutputDebugString( "Going from long name: " );
				OutputDebugString( fname.c_str() );
				OutputDebugString( "\n" );
				OutputDebugString( "to short name: ");
				OutputDebugString( short_fname );
				OutputDebugString( "\n" );
				fname = short_fname;
			}
		}
			
		if ( toolServices->IsFileOpen( fname ) )
		{ // only worry about it if it is opened in the editor
			bool mod = false;

			TIModuleInterface* mi = toolServices->GetModuleInterface( fname ); 

			// getting mi? this is changed in bcb
			if ( mi ) 
			{ // get editor interface for later checking
				OutputDebugString( "File " );
				OutputDebugString( fname.c_str() );
				OutputDebugString( " is opened.\n" );  
				
				TIEditorInterface* ei = mi->GetEditorInterface();
				if ( ei )
				{ // check for modified buffer
					OutputDebugString( "Got EditorInterface\n" );
					if ( ei->BufferModified() )
					{ // show warning and quit
						OutputDebugString( "Buffer Modified\n" );
						mod = true;
					}

					ei->Release();
				}

				if ( !mod )
				{ // double check form
					TIFormInterface* fi = mi->GetFormInterface();
					if ( fi )
					{ // check for modified form
						OutputDebugString( "Got Form Interface\n" );
						if ( fi->FormModified() )
						{
							OutputDebugString( "Form modified\n" );
							mod = true;
						}

						fi->Release();
					}
				}

				if ( !mod )
				{ // check aux
					TIEditorInterface* aei = mi->GetAuxEditorInterface();
					if ( aei )
					{ // check for modified form
						OutputDebugString( "Got Aux Editor Interface\n" );
						if ( aei->BufferModified() )
						{
							OutputDebugString( "Aux Editor modified\n" );
							mod = true;
						}

						aei->Release();
					}
				}

				// Tell IDE we're done with it
				mi->Release();
			}

			if ( mod )
				return;
			
			if ( !toolServices->ReloadFile( fname ) )
			{ 
				OutputDebugString( "\n###AnsiString Reload of " );
				OutputDebugString( fname.c_str() );
				OutputDebugString( " failed and I don't know why\n");
			}
		}
		else
		{ // doesn't look like it is loaded
			OutputDebugString( fname.c_str() );
			OutputDebugString( " is not currently loaded.\n" );
		}
	}
	else if ( strncmp( msg, "front", 5 ) == 0 )
	{ // wants IDE to come to the front
		HWND hWnd = toolServices->GetParentHandle();

		if (hWnd)
		{ // got window, bring it to the front
			if (IsIconic (hWnd))
			{
				ShowWindow (hWnd, SW_SHOWNORMAL);
					
				// SendMessage (hWnd, WM_SYSCOMMAND, SC_RESTORE, 0);
			}
			else
			{
				SetForegroundWindow(hWnd);
			}
		}
	}
			
}

void __stdcall TEmacsModuleNotifier::Notify( TNotifyCode nc )
{
	OutputDebugString( "Module Notify\n" );
	
	switch ( nc )
	{
	case ncEditorSelected:
		if ( !filename.Pos( ".c" ) && !filename.Pos( ".h" ) && !filename.Pos(".p") )
			return;										// Don't open non-source file

		ee->OpenFile( AnsiString(filename.c_str()) );
//		Sleep(500);
		break;
	}
}
		
#if DO_NOTIFIER
bool __stdcall setupModule(void * Param, const System::AnsiString FileName,
													 const System::AnsiString 	UnitName,
													 const System::AnsiString FormName)
{ 
	TIModuleInterface* mi = toolServices->GetModuleInterface( FileName );

	OutputDebugString( "setupModule: FileName= " );
	OutputDebugString( FileName.c_str() );
	OutputDebugString( "\nUnitName= " );
	OutputDebugString( UnitName.c_str() );
	OutputDebugString( "\nFormName= " );
	OutputDebugString( FormName.c_str() );
	OutputDebugString( "\n" );
	
	if ( !mi )
		if ( FormName.Length() == 0 ||
				 (mi = toolServices->GetFormModuleInterface( FormName ) ) == NULL )
			return true;

	OutputDebugString( "got mi\n\n");
	
	TEmacsModuleNotifier* emn = new TEmacsModuleNotifier();
	if ( emn )
	{
		emn->SetFileName( FileName );
		mi->AddNotifier( emn );
	}

	return true;
}
#endif

void __stdcall TEmacsNotifier::FileNotification(TFileNotification NotifyCode,
																								const System::AnsiString FileName,
																								bool &Cancel)
{
	{
		char buf[111];
		sprintf(buf, "FileNotification: %s - %d\n", FileName.c_str(), NotifyCode );
		
		OutputDebugString( buf );
	}
	
	switch (NotifyCode)
	{
	case fnFileOpening:
		OutputDebugString( "fnFileOpening\n" );
		break;
	case fnFileOpened:
		OutputDebugString( "fnFileOpened\n" );

		if ( !FileName.Pos( ".c" ) && !FileName.Pos( ".h" ) && !FileName.Pos(".p") )
			return;										// Don't open non-source file

		ee->OpenFile( AnsiString(FileName.c_str()) );
		break;
	case fnFileClosing:
		OutputDebugString( "fnFileClosing\n" );
		break;
	case fnProjectOpening:
		OutputDebugString( "fnProjectOpening\n" );
		break;
	case fnProjectOpened:
		OutputDebugString( "fnProjectOpened\n" );

#if DO_NOTIFIER
		toolServices->EnumProjectUnits( setupModule, 0 );
#endif
		
		break;
	case fnProjectClosing:
		OutputDebugString( "fnProjectClosing\n" );
		break;
	case fnAddedToProject:
		OutputDebugString( "fnAddedToProject\n" );
		break;
	case fnRemovedFromProject:
		OutputDebugString( "fnRemovedFromProject\n" );
		break;
	case fnDefaultDesktopLoad:
		OutputDebugString( "fnDefaultDesktopLoad\n" );
		break;
	case fnDefaultDesktopSave:
		OutputDebugString( "fnDefaultDesktopSave\n" );
		break;
	case fnProjectDesktopLoad:
		OutputDebugString( "fnProjectDesktopLoad\n" );
		break;
	case fnprojectDesktopSave:
		OutputDebugString( "fnprojectDesktopSave\n" );
		break;
	case fnPackageInstalled:
		OutputDebugString( "fnPackageInstalled\n" );
		break;
	case fnPackageUninstalled:
		OutputDebugString("fnPackageUninstalled\n");
		break;
	}
		 
	if ( fnProjectOpened == NotifyCode )
	{ // Done opening project, set up module notifiers
	}
	
#if 0
		if ( !projectOpening )
		{ // bring to front
			HWND hWnd = FindWindow ("Emacs", NULL);
			if (hWnd) {
				// Is the Emacs window iconified ?
				if (IsIconic (hWnd)) {
					ShowWindow (hWnd, SW_SHOWNORMAL);
					// Need this since Emacs thinks it is still iconified otherwise!
					SendMessage (hWnd, WM_SYSCOMMAND, SC_RESTORE, 0);
				}
				else {
					SetForegroundWindow(hWnd);
				}
			}
		}
#endif

}

// Standard IDE stuff


//---------------------------------------------------------------
extern "C" __declspec(dllexport) void __fastcall DoneExpert(void)
{
  // Put any general destruction code here.  Note that the
  // IDE will destroy any experts which have been registered.
}

bool InitEE( TIToolServices *ToolServices,
						 TExpertRegisterProc RegisterProc,
						 TExpertTerminateProc &Terminate)
{
	OutputDebugString( "INITEXPERT0017\n" );
	
  if (ToolServices == NULL)
    return false;

	toolServices = ToolServices;

	OutputDebugString( "Creating TEmacsEdit\n" );
	
	ee = new TEmacsEdit();
	
	OutputDebugString( "Checking for valid..\n");
	
	if ( ee && ee->IsValid() )
	{ // register it
		OutputDebugString( "ee is valid\n");
		Terminate = DoneExpert;
	
		(*RegisterProc)(ee);
		return true;
	}

	// delete and return false
	OutputDebugString( "ee is *NOT* valid\n" );
	delete ee;
	return false;
} // end of InitExpert()

//---------------------------------------------------------------
extern "C" __declspec(dllexport) bool _stdcall INITEXPERT0017(
    TIToolServices *ToolServices, TExpertRegisterProc RegisterProc,
    TExpertTerminateProc &Terminate)
{
	return InitEE( ToolServices, RegisterProc, Terminate );
}

extern "C" __declspec(dllexport) bool _stdcall INITEXPERT0016(
    TIToolServices *ToolServices, TExpertRegisterProc RegisterProc,
    TExpertTerminateProc &Terminate)
{
	return InitEE( ToolServices, RegisterProc, Terminate );
}

//---------------------------------------------------------------
